const std = @import("std");
const ascii = std.ascii;
const mem = std.mem;
const io = std.io;

pub const IniTokenizerVariant = enum {
    no_semicolon,
    semicolon,
};

pub const Token = union(enum) {
    section: []const u8,
    assign: struct {
        variable: []const u8,
        value: []const u8,
    },
};

pub fn tokenize(reader: anytype, comptime variant: IniTokenizerVariant) IniTokenizer(@TypeOf(reader), variant) {
    return .{ .reader = reader };
}

fn IniTokenizer(comptime T: type, comptime variant: IniTokenizerVariant) type {
    return struct {
        const Self = @This();

        reader: T,
        buffer: [4 * 1024]u8 = undefined,

        pub fn next(self: *Self, line_index: ?*usize) !?Token {
            while (true) {
                if (line_index) |i| i.* += 1;
                if (try self.reader.readUntilDelimiterOrEof(&self.buffer, '\n')) |raw_line| {
                    const line = mem.trim(u8, raw_line, &ascii.whitespace);
                    if (line.len == 0) continue;
                    if (line[0] == '#') continue;

                    // Line may contain comment after content.
                    const content = blk: {
                        for (line[1..], 0..) |ch, index| {
                            if (ch == '#' and line[index - 1] != '\\') {
                                break :blk mem.trim(u8, line[0 .. index + 1], &ascii.whitespace);
                            }
                        }
                        break :blk line;
                    };

                    // Is this line a section header?
                    if (content[0] == '[') {
                        if (content[content.len - 1] != ']') return error.InvalidLine;
                        if (content.len < 3) return error.InvalidLine;
                        return Token{ .section = content[1 .. content.len - 1] };
                    }

                    // Is this line an assignment?
                    const eql_index = mem.indexOfScalar(u8, content, '=') orelse return error.InvalidLine;
                    if (variant == .semicolon and eql_index == content.len - 1) return error.InvalidLine;
                    const variable = mem.trim(u8, content[0..eql_index], &ascii.whitespace);
                    if (variable.len == 0) return error.InvalidLine;
                    const value_quoted = blk: {
                        const value = mem.trim(u8, content[eql_index + 1 ..], &ascii.whitespace);
                        switch (variant) {
                            // In no_semicolon mode we explicitly allow empty lines.
                            .no_semicolon => break :blk value,
                            .semicolon => {
                                if (value.len < 2 or value[value.len - 1] != ';') return error.InvalidLine;
                                break :blk value[0 .. value.len - 1];
                            },
                        }
                    };
                    const value = blk: {
                        if (variant == .no_semicolon and value_quoted.len == 0) break :blk value_quoted;
                        inline for ([_]u8{ '\'', '"' }) |q| {
                            if (value_quoted[0] == q) {
                                if (value_quoted[value_quoted.len - 1] == q and value_quoted.len > 2) {
                                    break :blk value_quoted[1 .. value_quoted.len - 1];
                                }
                                return error.InvalidLine;
                            }
                        }
                        break :blk value_quoted;
                    };
                    return Token{ .assign = .{ .variable = variable, .value = value } };
                } else {
                    return null;
                }
            }
        }
    };
}

test "ini tokenizer good input" {
    var fbs = io.fixedBufferStream(
        \\[header] # I am a comment
        \\a = b;
        \\
        \\c=d;
        \\e= f;# Another comment
        \\
        \\[header2]
        \\[header3]
        \\#
        \\
        \\hello = this has spaces;
        \\hello = this one; is weird;
        \\hello = test=test;
        \\
    );
    const reader = fbs.reader();

    var it = tokenize(reader, .semicolon);
    var line: usize = 0;

    if (try it.next(&line)) |a| {
        try std.testing.expect(a == .section);
        try std.testing.expectEqualSlices(u8, "header", a.section);
        try std.testing.expect(line == 1);
    } else {
        unreachable;
    }

    if (try it.next(&line)) |a| {
        try std.testing.expect(a == .assign);
        try std.testing.expectEqualSlices(u8, "a", a.assign.variable);
        try std.testing.expectEqualSlices(u8, "b", a.assign.value);
        try std.testing.expect(line == 2);
    } else {
        unreachable;
    }

    if (try it.next(&line)) |a| {
        try std.testing.expect(a == .assign);
        try std.testing.expectEqualSlices(u8, "c", a.assign.variable);
        try std.testing.expectEqualSlices(u8, "d", a.assign.value);
        try std.testing.expect(line == 4);
    } else {
        unreachable;
    }

    if (try it.next(&line)) |a| {
        try std.testing.expect(a == .assign);
        try std.testing.expectEqualSlices(u8, "e", a.assign.variable);
        try std.testing.expectEqualSlices(u8, "f", a.assign.value);
        try std.testing.expect(line == 5);
    } else {
        unreachable;
    }

    if (try it.next(&line)) |a| {
        try std.testing.expect(a == .section);
        try std.testing.expectEqualSlices(u8, "header2", a.section);
        try std.testing.expect(line == 7);
    } else {
        unreachable;
    }

    if (try it.next(&line)) |a| {
        try std.testing.expect(a == .section);
        try std.testing.expectEqualSlices(u8, "header3", a.section);
        try std.testing.expect(line == 8);
    } else {
        unreachable;
    }

    if (try it.next(&line)) |a| {
        try std.testing.expect(a == .assign);
        try std.testing.expectEqualSlices(u8, "hello", a.assign.variable);
        try std.testing.expectEqualSlices(u8, "this has spaces", a.assign.value);
        try std.testing.expect(line == 11);
    } else {
        unreachable;
    }

    if (try it.next(&line)) |a| {
        try std.testing.expect(a == .assign);
        try std.testing.expectEqualSlices(u8, "hello", a.assign.variable);
        try std.testing.expectEqualSlices(u8, "this one; is weird", a.assign.value);
        try std.testing.expect(line == 12);
    } else {
        unreachable;
    }

    if (try it.next(&line)) |a| {
        try std.testing.expect(a == .assign);
        try std.testing.expectEqualSlices(u8, "hello", a.assign.variable);
        try std.testing.expectEqualSlices(u8, "test=test", a.assign.value);
        try std.testing.expect(line == 13);
    } else {
        unreachable;
    }

    try std.testing.expect((try it.next(&line)) == null);
}

test "ini tokenizer bad input" {
    {
        var fbs = io.fixedBufferStream(
            \\[section
            \\
        );
        const reader = fbs.reader();
        var it = tokenize(reader, .semicolon);
        var line: usize = 0;
        try std.testing.expectError(error.InvalidLine, it.next(&line));
        try std.testing.expect(line == 1);
    }
    {
        var fbs = io.fixedBufferStream(
            \\section]
            \\
        );
        const reader = fbs.reader();
        var it = tokenize(reader, .semicolon);
        var line: usize = 0;
        try std.testing.expectError(error.InvalidLine, it.next(&line));
        try std.testing.expect(line == 1);
    }
    {
        var fbs = io.fixedBufferStream(
            \\[]
            \\
        );
        const reader = fbs.reader();
        var it = tokenize(reader, .semicolon);
        var line: usize = 0;
        try std.testing.expectError(error.InvalidLine, it.next(&line));
        try std.testing.expect(line == 1);
    }
    {
        var fbs = io.fixedBufferStream(
            \\ =B;
            \\
        );
        const reader = fbs.reader();
        var it = tokenize(reader, .semicolon);
        var line: usize = 0;
        try std.testing.expectError(error.InvalidLine, it.next(&line));
        try std.testing.expect(line == 1);
    }
    {
        var fbs = io.fixedBufferStream(
            \\a =
            \\
        );
        const reader = fbs.reader();
        var it = tokenize(reader, .semicolon);
        var line: usize = 0;
        try std.testing.expectError(error.InvalidLine, it.next(&line));
        try std.testing.expect(line == 1);
    }
    {
        var fbs = io.fixedBufferStream(
            \\a =  ;
            \\
        );
        const reader = fbs.reader();
        var it = tokenize(reader, .semicolon);
        var line: usize = 0;
        try std.testing.expectError(error.InvalidLine, it.next(&line));
        try std.testing.expect(line == 1);
    }
}
